From 613e92cefe3f8cdfbb2388c1f66be76a79d6c3a9 Mon Sep 17 00:00:00 2001
From: Min Xu <min.xu[A.T]min-info[D.O.T]net>
Date: Mon, 7 Oct 2013 08:32:33 -1000
Subject: [PATCH 6/6] Retrive previous PC from the stack before entering
 specific interrupt handler routines.  Allow user to
 trace where interrupt occured: e.g. WatchDog and
 Spurious interrupt

Prior to this change, spurious interrupt would occur so much (observed
via gdb/remote debug) so it appears that the board stalled.  Once a
custom spurious interrupt installed, the code continues after the
interrupt instead of re-entering the interrupt.
---
 firmware/src/os/system_irq.c          |   17 ++++++++++++++---
 firmware/src/simtrace/main_simtrace.c |   16 +++++++++++++++-
 2 files changed, 29 insertions(+), 4 deletions(-)

diff --git a/firmware/src/os/system_irq.c b/firmware/src/os/system_irq.c
index dc787eb..cc1fa08 100644
--- a/firmware/src/os/system_irq.c
+++ b/firmware/src/os/system_irq.c
@@ -25,15 +25,15 @@
 
 #include <os/system_irq.h>
 #include <os/dbgu.h>
+#include <string.h>
 
 #include "../openpcd.h"
 
 static sysirq_hdlr *sysirq_hdlrs[AT91SAM7_SYSIRQ_COUNT];
 
-static void sys_irq(void)
+void sys_irq(u_int32_t previous_pc)
 {
 	u_int32_t sr;
-	DEBUGP("sys_irq ");
 
 	/* Somehow Atmel decided to do really stupid interrupt sharing
 	 * for commonly-used interrupts such as the timer irq */
@@ -139,6 +139,10 @@ static void sys_irq(void)
 	if (*AT91C_WDTC_WDMR & AT91C_WDTC_WDFIEN) {
 		sr = *AT91C_WDTC_WDSR;
 		if (sr) {
+			char dbg_buf[100];
+			sprintf(dbg_buf, "sys_irq [Old PC = %08X]\n\r", previous_pc);
+			AT91F_DBGU_Frame(dbg_buf);
+
 			DEBUGP("WDT(");
 			if (sysirq_hdlrs[AT91SAM7_SYSIRQ_WDT]) {
 				DEBUGP("handler ");
@@ -154,6 +158,13 @@ static void sys_irq(void)
 	DEBUGPCR("END");
 }
 
+static void sysirq_entry() {
+  	// DO NOT MODIFY THIS SECTION AND Cstartup.S/IRQ_Handler_Entry.
+	register unsigned *previous_pc asm("r0");
+	asm("ADD R1, SP, #16; LDR R0, [R1]"); 
+	sys_irq(previous_pc);
+}
+
 void sysirq_register(enum sysirqs irq, sysirq_hdlr *hdlr)
 {
 	if (irq >= AT91SAM7_SYSIRQ_COUNT)
@@ -167,6 +178,6 @@ void sysirq_init(void)
 	AT91F_AIC_ConfigureIt(AT91C_BASE_AIC, AT91C_ID_SYS,
 			      OPENPCD_IRQ_PRIO_SYS,
 			      AT91C_AIC_SRCTYPE_INT_HIGH_LEVEL,
-			      &sys_irq);
+			      &sysirq_entry);
 	AT91F_AIC_EnableIt(AT91C_BASE_AIC, AT91C_ID_SYS);
 }
diff --git a/firmware/src/simtrace/main_simtrace.c b/firmware/src/simtrace/main_simtrace.c
index edbd9b4..137f188 100644
--- a/firmware/src/simtrace/main_simtrace.c
+++ b/firmware/src/simtrace/main_simtrace.c
@@ -136,8 +136,22 @@ static int simtrace_usb_in(struct req_ctx *rctx)
 	}
 }
 
+void custom_spurious_handler(unsigned previous_pc) {
+	char dbg_buf[100];
+	sprintf(dbg_buf, "SPURRIOUS IRQ [Old PC = %08X]\n\r", previous_pc);
+	AT91F_DBGU_Frame(dbg_buf);
+} 
+
+void custom_spurious_entry(void) {
+	register unsigned *previous_pc asm("r0");
+	asm("ADD R1, SP, #16; LDR R0, [R1]");
+	custom_spurious_handler(previous_pc);
+}
+
 void _init_func(void)
 {
+	AT91C_BASE_AIC->AIC_SPU = (int)custom_spurious_entry;
+
 	/* low-level hardware initialization */
 	pio_irq_init();
 	iso_uart_init();
@@ -231,7 +245,7 @@ void _main_func(void)
        	if ((loopLow & 0xFFFF) == 0) {
 		DEBUGPCR("Heart beat %08X", loopHigh++);
 	}
-	if ((loopLow & 0xFFF) == 0) {
+	if ((loopLow & 0x3FF) == 0) {
 		iso_uart_idleflush();
 	}
 	loopLow++;
-- 
1.7.9.5

